from .arguments import Argument
import argparse
import sys
from collections import OrderedDict
# from pprint import pprint


class BaseMeta(type):
    @classmethod
    def __prepare__(mcls, cls, bases):
        return OrderedDict()


class CLIMeta(BaseMeta):
    def __new__(cls, name, bases, attrs):
        obj = super().__new__(cls, name, bases, attrs)
        obj._meta = type('CLIMeta', (object, ), {
            'command': True,
            'arguments': obj.__get_arguments(attrs),
            'bases': bases,
        })
        return obj

    def __get_arguments(self, attrs):
        # print(type(self), attrs)
        return OrderedDict([(argname, arg) for argname, arg in attrs.items() if isinstance(arg, Argument)])


class TieredCLIMeta(BaseMeta):
    def __new__(cls, name, bases, attrs):
        obj = super().__new__(cls, name, bases, attrs)
        obj._meta = type('TieredCLIMeta', (object, ), {
            'command': False,
            'subcommands': obj.__get_subcommands(attrs, bases),
            'bases': bases,
        })
        return obj

    def __get_subcommands(self, attrs, bases):
        subcommands = []
        qualnames = [base.__qualname__ for base in bases]

        if 'TieredCLI' not in qualnames:
            return None

        for argname, arg in attrs.items():
            if hasattr(arg, '_meta'):
                qualnames = [base.__qualname__ for base in arg._meta.bases]
                if 'CLI' in qualnames:
                    subcommands.append((argname, arg))

        return OrderedDict(subcommands)


class BaseCLI(object):
    __slots__ = ('_meta', 'call')

    def __init__(self):
        self._setup_parser()
        self._parse_args()

    def _setup_parser(self):
        assert False, "_setup_parser not defined"

    def _parse_args(self):
        self.parser.parse_args(namespace=self)


class CLI(BaseCLI, metaclass=CLIMeta):
    def _setup_parser(self):
        self.parser = argparse.ArgumentParser()

        for name, arg in self._meta.arguments.items():
            self.parser.add_argument(
                name,
                # required=arg.required,
                type=arg.argtype,
                help=arg.help_text,
            )


class TieredCLI(BaseCLI, metaclass=TieredCLIMeta):
    def _setup_parser(self):
        self.parser = argparse.ArgumentParser()

        self.parser.add_argument(
            'command',
            type=str,
        )

        self.parser.add_argument(
            'arguments',
            nargs='+',
        )

    async def call(self):
        cmd = self._meta.subcommands.get(self.command, None)

        if cmd is None:
            raise Exception('Could not find command {}'.format(self.command))

        del sys.argv[1]

        command = cmd()
        await command.call()



import asynccli
from unittest.mock import patch


class Calculator(asynccli.CLI):
    first_num = asynccli.Integer(help_text='This is some help text.')
    second_num = asynccli.Integer()

    async def call(self):
        print(self.first_num / self.second_num)


def test_division_calculator(capsys):
    # sys.argv = ['1', '2']

    with patch('sys.argv', ['', '1', '2']):
        app = asynccli.App(Calculator)
        app.run()

    out, _ = capsys.readouterr()
    assert out == "0.5\n"



import asyncio
from .cli import BaseCLI


class App(object):
    def __init__(self, *tasks):
        self.loop = asyncio.get_event_loop()
        self.tasks = tasks

    def run(self):
        tasks = self.__call_tasks()
        self.loop.run_until_complete(asyncio.wait(tasks))
        # self.loop.close()

    def __call_tasks(self):
        tasks = []
        for task in self.tasks:
            called = task()
            if isinstance(called, BaseCLI):
                tasks.append(called.call())
            else:
                tasks.append(called)

        return tasks



class Argument(object):
    __slots__ = ('argtype', 'required', 'help_text')

    def __init__(
        self,
        required=False,
        help_text=None,
    ):
        self.required = required
        self.help_text = help_text


class Integer(Argument):
    argtype = int



"""asynccli - A CLI framework based on asyncio"""

from .app import App
from .cli import CLI, TieredCLI, TieredCLIMeta
from .arguments import *

__version__ = '0.1.0'
__author__ = 'Adam Hopkins <admhpkns@gmail.com>'
__all__ = [
    'App',
    'CLI',
    'TieredCLI',
    'TieredCLIMeta',
    'Integer',
]



import asynccli


async def mycli():
    print("Hello, world.")


class MyCLI(asynccli.CLI):
    async def call(self):
        print("Hello, world.")


def test_hello_world_basic(capsys):
    app = asynccli.App(mycli)
    app.run()

    out, _ = capsys.readouterr()
    assert out == "Hello, world.\n"


def test_hello_world_class(capsys):
    app = asynccli.App(MyCLI)
    app.run()

    out, _ = capsys.readouterr()
    assert out == "Hello, world.\n"



import setuptools

setuptools.setup(
    name="asynccli",
    version="0.1.1",
    url="https://github.com/ahopkins/asynccli",

    author="Adam Hopkins",
    author_email="admhpkns@gmail.com",

    description="A CLI framework based on asyncio",
    long_description=open('README.rst').read(),

    packages=setuptools.find_packages(),

    install_requires=[],

    setup_requires=['pytest-runner', ],
    tests_require=['pytest', ],

    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



import asynccli
from unittest.mock import patch


class DivisionCalculator(asynccli.CLI):
    first_num = asynccli.Integer(help_text='This is some help text.')
    second_num = asynccli.Integer()

    async def call(self):
        print(self.first_num / self.second_num)


class MultiplicationCalculator(asynccli.CLI):
    first_num = asynccli.Integer(help_text='This is some help text.')
    second_num = asynccli.Integer()

    async def call(self):
        print(self.first_num * self.second_num)


class Calculator(asynccli.TieredCLI):
    d = DivisionCalculator
    m = MultiplicationCalculator


def test_calculator_with_multiplication(capsys):
    with patch('sys.argv', ['', 'm', '2', '4']):
        app = asynccli.App(Calculator)
        app.run()

    out, _ = capsys.readouterr()
    assert out == "8\n"


def test_calculator_with_division(capsys):
    with patch('sys.argv', ['', 'd', '2', '4']):
        app = asynccli.App(Calculator)
        app.run()

    out, _ = capsys.readouterr()
    assert out == "0.5\n"
